#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <omp.h>

#include "measures.h"

#define min(a,b) ((a<b) ? (a) : (b))

// float uses cache better
typedef double FLOAT;
//typedef float FLOAT;

//------------------------------------------------------------------------
// vanilla variant
//------------------------------------------------------------------------

// allocate memory for matrices
void allocMatrixVanilla(FLOAT*** A, int n)
{
  int i, j;

  // allocate row memory
  *A = (FLOAT**) malloc(n*sizeof(FLOAT*));

  // allocate column memory
  for (i=0; i<n; i++)
    (*A)[i] = (FLOAT*) malloc(n*sizeof(FLOAT));

  // set to zero
  for(i=0; i<n; i++)
    for(j=0; j<n; j++)
      (*A)[i][j] = 0.0;
}

// initialize matrix
void initMatrixVanilla(FLOAT** A, int n)
{
  int i, j;

  for(i=0; i<n; i++)
    for(j=0; j<n; j++)
      A[i][j] = (FLOAT) (i+j);
}

// release memory
void freeMatrixVanilla(FLOAT** A, int n)
{
  int i;

  // release rows first then columns
  for (i=0; i<n; i++)
    free(A[i]);
  free(A);

  // secure pointer
  A = NULL;
}

// repeated naive matrix multiplication
inline void matrixMultVanilla(FLOAT** A, FLOAT** B, FLOAT** C, int n, int repeat)
{
  int i, j, k, r;

  for(r=0; r<repeat; r++)
    for(i=0; i<n; i++)
      for(j=0; j<n; j++)
        for(k=0; k<n; k++)
          C[i][j] += A[i][k] * B[k][j];
}

inline void runVanilla(int n, int r)
{
  // pointers to matrix structures
  FLOAT** A = NULL;
  FLOAT** B = NULL;
  FLOAT** C = NULL;

  // allocate memory
  allocMatrixVanilla(&A, n);
  allocMatrixVanilla(&B, n);
  allocMatrixVanilla(&C, n);

  // init matrices
  initMatrixVanilla(A, n);
  initMatrixVanilla(B, n);

  // run simulation with time measurement
  double tstart = omp_get_wtime();
  matrixMultVanilla(A, B, C, n, r);
  double tend = omp_get_wtime();
  const double tt = tend-tstart;

  // analyze results (n, MFLOPs, t_proc, repeats)
  printf("%-6i %8.2f %10.4f %12i\n", n, r*mflop(n)/tt, tt, r);
  fflush(stdout);

  // free memory
  freeMatrixVanilla(A, n);
  freeMatrixVanilla(B, n);
  freeMatrixVanilla(C, n);

  return;
}

//------------------------------------------------------------------------
// tiled variant
//------------------------------------------------------------------------

// initialize block matrices
inline void setMatrixTiled(FLOAT* A, int n, int T, int i, int j, FLOAT v)
{
  // first summand: find block; second summand: find position in block
  int x = T * T * ((i/T)* (n/T) + (j/T)) + ((i%T) * T + (j%T));
  A[x] = v;
}

// initialize block matrix
void initMatrixTiled(FLOAT* A, int n, int T)
{
  int i, j;

  for(i=0; i<n; i++)
    for(j=0; j<n; j++)
      setMatrixTiled(A, n, T, i, j, (FLOAT)(i+j));
}

// allocate memory for block matrix
void allocMatrixTiled(FLOAT** A, int n, int T)
{
  *A = (FLOAT*) malloc(n * n * sizeof(FLOAT));
}

// release memory of block matrix
void freeMatrixTiled(FLOAT* A, int n, int T)
{
  free(A);
}

// please doc me
inline void matrixMultLoopTiled(FLOAT* A, FLOAT* B, FLOAT* C, int n, int T, int repeat)
{
  int i, j, k, s, t, u, r;

  /* block indices */
  int bi = T/4;
  int bj = T;
  int bk = T*4;

  for (r=0; r<repeat; r++)
  {
    // block loop
    for (i=0; i<n; i+=bi)
    {
      for (j=0; j<n; j+=bj)
      {
        for (k=0; k<n; k+=bk)
        {
          // loop in single block
          for (s=i; s<min(i+bi, n); s++)
          {
            for (t=j; t<min(j+bj, n); t++)
            {
              for (u=k; u<min(k+bk, n); u++)
              {
                // C = A * B
                C[s*n+t] += A[s*n+u] * B[u*n+t];
              }
            }
          }
        }
      }
    }
  }

  return;
}

// run tiled */
void runTiled(int n, int repeat, int T)
{
  FLOAT* A = NULL;
  FLOAT* B = NULL;
  FLOAT* C = NULL;

  // allocate storage for full matrix
  allocMatrixTiled(&A, n, 1);
  allocMatrixTiled(&B, n, 1);
  allocMatrixTiled(&C, n, 1);

  // initialize full matrix
  initMatrixTiled(A, n, 1);
  initMatrixTiled(B, n, 1);

  double tstart = omp_get_wtime();
  matrixMultLoopTiled(A, B, C, n, T, repeat);
  double tend = omp_get_wtime();
  const double tt = tend - tstart;

  // analyze results (n, MFLOPs, t_proc, repeats, tilesize)
  printf("%-6i %8.2f %10.4f %-8i %-8i\n", n, repeat*mflop(n)/tt, tt, repeat, T);
  fflush(stdout);

  // release memoryi
  freeMatrixTiled(A, n, 1);
  freeMatrixTiled(B, n, 1);
  freeMatrixTiled(C, n, 1);
}

//------------------------------------------------------------------------
// main
//------------------------------------------------------------------------
int main(int argc, char** argv)
{
  int n = 1; // matrix row and column size
  int r = 1; // # repeats

  if (argc != 3)
  {
    printf("Usage: %s <Matrix row size N> <repeats>\n", argv[0]);
    exit(1);
  }
  else
  {
    n = atoi(argv[1]);
    r = atoi(argv[2]);
  }

  // vanilla (untiled) variant
  printf("n        FLOPs      tElapsed    repeats\n");
  runVanilla(n, r);

  // tiled variant, t >= 2!
  //int t;
  //printf("n    FLOPs    tElapsed    repeats    tilesize\n");
  //for (t=2; t<8; ++t)
  //{
  //  runTiled(n, r, 2<<t);
  //}

  return 0;
}
